﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO;
using Microsoft.Win32;
using Microsoft.Win32.SafeHandles;
using System.Runtime.InteropServices;

namespace geonamesThreads
{
    unsafe public class FileMm : IDisposable
    {
        private const uint PAGE_READONLY = 0x00000002;
        private const uint PAGE_READWRITE = 0x00000004;
        private const uint FILE_MAP_READ = 0x00000004;
        private const uint FILE_MAP_ALL_ACCESS = 0x00000002;

        [DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern SafeFileHandle CreateFile(String lpFileName,
            FileAccess dwDesiredAccess, FileShare dwShareMode,
            IntPtr lpSecurityAttributes, FileMode dwCreationDisposition,
            uint dwFlagsAndAttributes, IntPtr hTemplateFile);

        [DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern bool CloseHandle(SafeFileHandle h);

        [DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern bool SetFilePointerEx(SafeFileHandle h, Int64 Distance, Int64* pPosition, uint dwMoveMethod);

        [DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern bool SetEndOfFile(SafeFileHandle h);

        [DllImport("kernel32.dll", SetLastError = true)]
        private static extern SafeFileHandle CreateFileMapping(SafeFileHandle hFile,
            IntPtr lpSecurityAttributes, UInt32 dwProtect, UInt32 dwMaximumSizeHigh,
            UInt32 dwMaximumSizeLow, IntPtr lpMapName);

        [DllImport("kernel32.dll", SetLastError = true)]
        private static extern byte* MapViewOfFile(SafeFileHandle hMapObject,
            UInt32 dwAccess, Int32 dwOffsetHigh, UInt32 dwOffsetLow, UInt64 dwNumberOfBytesToMap);

        [DllImport("kernel32.dll", SetLastError = true)]
        private static extern bool UnmapViewOfFile(byte* lpBaseAddress);

        [DllImport("kernel32.dll", SetLastError = true)]
        private static extern bool FlushViewOfFile(byte* lpBaseAddress, UInt64 dwNumberOfBytesToFlush);

        [DllImport("kernel32.dll", SetLastError = true)]
        private static extern int GetLastError();

        /// <summary>
        /// Mmeory Mapped Read Only file access constructor.
        /// </summary>
        /// <param name="inFile">Input file path name</param>

        public FileMm(String inFile)
        {
            _inHandle = CreateFile(inFile, FileAccess.Read, FileShare.Read, IntPtr.Zero, FileMode.Open, 0, IntPtr.Zero);
            if (_inHandle.IsInvalid) throw new Exception("Error. Failed to open input file " + inFile);

            FileInfo fi = new FileInfo(inFile);
            _FileSize = (UInt64)fi.Length;

            SafeFileHandle inMapHandle = CreateFileMapping(_inHandle, IntPtr.Zero, PAGE_READONLY, 0, 0, IntPtr.Zero);
            if (inMapHandle == null)
            {
                throw new Exception("Failed to CreateFileMapping for input file mapping.");
            }
            _pIn = (byte*)MapViewOfFile(inMapHandle, FILE_MAP_READ, 0, 0, (UInt64)0); CloseHandle(inMapHandle);
            if (_pIn == null) throw new Exception("Error. Failed to map view of input file.");
        }

        private UInt64 _FileSize;
        private byte* _pIn = null;
        private SafeFileHandle _inHandle = null;

        public UInt64 FileSize { get { return _FileSize; } }
        public byte* pIn { get { return _pIn; } }

        public void Dispose()
        {
            if (pIn != null) UnmapViewOfFile(pIn);
            if (!_inHandle.IsInvalid) CloseHandle(_inHandle);
        }
    }
}
