﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO;
using Microsoft.Win32;
using Microsoft.Win32.SafeHandles;
using System.Runtime.InteropServices;

namespace MemoryMappedNative
{
    unsafe public class FileMmChar : IEnumerable<String>
    {
        private const uint PAGE_READONLY = 0x00000002;
        private const uint PAGE_READWRITE = 0x00000004;
        private const uint FILE_MAP_READ = 0x00000004;
        private const uint FILE_MAP_ALL_ACCESS = 0x00000002;

        [DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern SafeFileHandle CreateFile(String lpFileName,
            FileAccess dwDesiredAccess, FileShare dwShareMode,
            IntPtr lpSecurityAttributes, FileMode dwCreationDisposition,
            uint dwFlagsAndAttributes, IntPtr hTemplateFile);

        [DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern bool CloseHandle(SafeFileHandle h);

        [DllImport("kernel32.dll", SetLastError = true)]
        private static extern SafeFileHandle CreateFileMapping(SafeFileHandle hFile,
            IntPtr lpSecurityAttributes, UInt32 dwProtect, UInt32 dwMaximumSizeHigh,
            UInt32 dwMaximumSizeLow, IntPtr lpMapName);

        [DllImport("kernel32.dll", SetLastError = true)]
        private static extern byte* MapViewOfFile(SafeFileHandle hMapObject,
            UInt32 dwAccess, Int32 dwOffsetHigh, UInt32 dwOffsetLow, UInt64 dwNumberOfBytesToMap);

        [DllImport("kernel32.dll", SetLastError = true)]
        private static extern bool UnmapViewOfFile(byte* lpBaseAddress);

        [DllImport("kernel32.dll", SetLastError = true)]
        private static extern int GetLastError();

        /// <summary>
        /// Mmeory Mapped file access constructor.
        /// </summary>
        /// <param name="pathName">File path name</param>
        /// <param name="readOnly">read only or read-write?</param>

        public FileMmChar(String pathName, bool readOnly)
        {
            _fileHandle = CreateFile(pathName, FileAccess.Read, readOnly ? FileShare.Read : FileShare.ReadWrite, IntPtr.Zero, FileMode.Open, 0, IntPtr.Zero);
            if (_fileHandle.IsInvalid) throw new Exception("Error. Failed to open file " + pathName + "Error: " + GetLastError().ToString());

            FileInfo fi = new FileInfo(pathName);
            _FileSize = fi.Length;

            SafeFileHandle inMapHandle = CreateFileMapping(_fileHandle, IntPtr.Zero, readOnly ? PAGE_READONLY : PAGE_READWRITE, 0, 0, IntPtr.Zero);
            if (inMapHandle == null) throw new Exception("Failed to CreateFileMapping for mapping for file " + pathName + "Error: " + GetLastError().ToString());

            _pFile = (byte*)MapViewOfFile(inMapHandle, readOnly ? FILE_MAP_READ : FILE_MAP_ALL_ACCESS, 0, 0, (UInt64)0); CloseHandle(inMapHandle);
            if (_pFile == null) throw new Exception("Error. Failed to map view of file " + pathName + "Error: " + GetLastError().ToString());
        }

        private Int64 _FileSize;
        private byte* _pFile = null;
        private SafeFileHandle _fileHandle = null;

        public Int64 FileSize { get { return _FileSize; } }
        public byte* pFile { get { return _pFile; } }

        /// <summary>
        /// IEnumerable<String> support. "Char" is the class name suffix as a String is a collection of Char objects.
        /// To be done. Combine this class and FileMmByte into a single generic class.
        /// </summary>
        /// <returns></returns>
        IEnumerator<String> IEnumerable<String>.GetEnumerator()
        {
            return new LineEnumeratorChar(this);
        }

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            throw new NotImplementedException();
        }

        public static IEnumerable<String> ReadLines(String path)
        {
            FileMmChar myFile = new FileMmChar(path, true);  // Read only
            IEnumerator<String> stringEnumerator = new LineEnumeratorChar(myFile);
            foreach (String line in myFile)
            {
                yield return line;
            }
            myFile.Dispose();
        }

        public void Dispose()
        {
            if (pFile != null) UnmapViewOfFile(pFile);
            if (!_fileHandle.IsInvalid) CloseHandle(_fileHandle);
        }

    }

}
