﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MemoryMappedNative
{
    unsafe class ByteLineEnumerator<TItem> : IEnumerator<byte[]>
    {
        /// <summary>
        /// Supports class FileMmChar IEnumerable<String> interface.
        /// </summary>
        public ByteLineEnumerator(FileMmByte mappedFile)
        {
            this.currentData = mappedFile;
            lineStart = -1;
            currentLength = 0;
            currentBegin = 0;
        }

        private FileMmByte currentData;
        private Int64 lineStart;
        private Int64 currentBegin;
        private int currentLength;

        byte[] IEnumerator<byte[]>.Current
        {
            get
            {
                if (this.lineStart < 0 || currentLength <= 0)
                    throw new InvalidOperationException("Use MoveNext before calling Current");
                byte[] result = new byte[this.currentLength];
                Int64 currentEnd = currentBegin + currentLength;
                Int64 i = currentBegin, j = 0;
                while (i < currentEnd)
                {
                    result[j++] = currentData.pFile[i++];
                }
                // ?? - could we use System.Array.Copy() for more performance?
                return result;
            }
        }

        void IDisposable.Dispose()
        {
        }

        object System.Collections.IEnumerator.Current
        {
            get { throw new NotImplementedException(); }
        }

        bool System.Collections.IEnumerator.MoveNext()
        {
            const byte lf = 0x0A;
            Int64 fileSizeMinus1 = currentData.FileSize - 1;
           
            if (lineStart > fileSizeMinus1) return false;

            if (this.lineStart < 0) this.lineStart = 0;  // First call to MoveNext
            currentBegin = lineStart;

            // Scan the byte[] currentData.pFile[i] from the line starting at file position lineStart and update lineStart.
            // The end of the line is identified by the new line (0x0A) value, not by the byte[] length. Retain the 
            // beginning position. do not move the data; just retain the line boundaries.
            byte c;
            for (Int64 i = lineStart; i <= fileSizeMinus1; i++)
            {
                c = currentData.pFile[i];
                if (c == lf)
                {
                    currentLength = (int)(i - lineStart);
                    lineStart = i + 1; // Prepare for next MoveNext() call
                    break;
                }
            }
            return true;
        }

        void System.Collections.IEnumerator.Reset()
        {
            lineStart = -1;
            currentLength = 0;
        }
    }
}
