﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO;
using System.Net;

namespace MemoryMappedNative
{
    /// <summary>
    /// Supports class FileMmChar IEnumerable<String> interface.
    /// </summary>
    unsafe class LineEnumeratorChar : IEnumerator<System.String>
    {
        public LineEnumeratorChar(FileMmChar mappedFile)
        {
            this.currentData = mappedFile;
        }

        private FileMmChar currentData = null;
        private System.String currentItem = default(System.String);
        private StringBuilder currentItemBuilder = new StringBuilder("", 10000);
        private System.Int64 lineStart = -1;
        private int lineNumber = -1;

        System.String IEnumerator<String>.Current
        {
            get
            {
                if (this.lineStart < 0)
                    throw new InvalidOperationException("Use MoveNext before calling Current");
                return this.currentItem;
            }
        }

        void IDisposable.Dispose()
        {
        }

        object System.Collections.IEnumerator.Current
        {
            get { throw new NotImplementedException(); }
        }

        bool System.Collections.IEnumerator.MoveNext()
        {
             Int64 fileSizeMinus1 = currentData.FileSize - 1;
            if (lineStart > fileSizeMinus1) return false;

            if (this.lineStart < 0) this.lineStart = 0;  // First call to MoveNext
            lineNumber++;

            // Create a String from the line starting at file position lineStart and update lineStart.
            byte c;

            currentItemBuilder.Clear();

            for (Int64 i = lineStart; i <= fileSizeMinus1; i++)
            {
                c = currentData.pFile[i];
                if (c != 0xA)
                {
                    currentItemBuilder.Append((Char)c);
                }
                else
                {
                    currentItem = currentItemBuilder.ToString();
                    lineStart = i + 1;
                    return true;
                }
            }
            return true;
        }

        void System.Collections.IEnumerator.Reset()
        {
            currentItem = default(System.String);
            lineStart = 0;
        }
    }
}
